#include "config.h"
#include "defines.h"
#include "mystring.h"
#include "list.h"
#include "hostname.h"
#include "fdbuf.h"
#include "itoa.h"
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

fdobuf fout(1);
fdibuf fin(0);

#define fail(MSG) do{ fout.write("nullmailer-inject: "); fout.writeln(MSG); return false; }while(0)
#define fail_sys(MSG) do{ fout.write("nullmailer-inject: "); fout.write(MSG); fout.write(": "); fout.writeln(strerror(errno)); return false; }while(0)

typedef list<mystring> slist;
static bool do_debug = false;

///////////////////////////////////////////////////////////////////////////////
// Envelope processing
///////////////////////////////////////////////////////////////////////////////
static slist recipients;
static mystring sender;
static bool use_header_recips = true;

mystring parse_address(mystring recip)
{
  recip = recip.strip();
  if(do_debug) {
    fout.write("Parsing address: ");
    fout.write(recip);
  }
  int i;
  int j;
  if((i = recip.find('<')) >= 0 &&
     (j = recip.find('>', i)) >= i)
    recip = recip.sub(i+1, j-i-1);
  else if((i = recip.find('(')) >= 0 &&
	  (j = recip.find(')', i)) >= i)
    recip = recip.sub(0, i-1).rstrip();
  if(do_debug) {
    fout.write("->");
    fout.writeln(recip);
  }
  return recip;
}

void parse_recip(mystring r)
{
  r = parse_address(r);
  if(!!r)
    recipients.append(r);
}

void parse_recips(const char* str)
{
  if(use_header_recips) {
  const char* start = str;
  const char* end;
  while((end = strchr(start, ',')) != 0) {
    parse_recip(mystring(start, end-start));
    start = end+1;
  }
  parse_recip(start);
  }
}

///////////////////////////////////////////////////////////////////////////////
// Header processing
///////////////////////////////////////////////////////////////////////////////
static slist headers;
static bool has_from = false;
static bool has_date = false;
static bool has_mid = false;

static mystring from;
static mystring host_name = hostname();

void set_from()
{
  const char* user = getenv("NULLMAILER_USER");
  if(!user) user = getenv("MAILUSER");
  if(!user) user = getenv("USER");
  if(!user) user = getenv("LOGNAME");
  if(!user) user = "unknown";

  const char* host = getenv("NULLMAILER_HOST");
  if(!host) host = getenv("MAILHOST");
  if(!host) host = getenv("HOSTNAME");
  if(!host) host = host_name.c_str();

  const char* name = getenv("NULLMAILER_NAME");
  if(!name) name = getenv("MAILNAME");
  if(!name) name = getenv("NAME");

  from = mystring(name, " <", user);
  from += mystring("@", host, ">");

  const char* suser = getenv("NULLMAILER_SUSER");
  if(!suser) suser = user;

  const char* shost = getenv("NULLMAILER_SHOST");
  if(!shost) shost = host;

  sender = mystring(suser, "@", shost);
}

void parse_line(mystring& line)
{
  const char* str = line.c_str();
  if(!strncasecmp(str, "from:", 5)) {
    mystring f = parse_address(str+5);
    if(!!f)
      sender = f;
    has_from = true;
  }
  else if(!strncasecmp(str, "date:", 5))
    has_date = true;
  else if(!strncasecmp(str, "message-id:", 11))
    has_mid = true;
  else if(!strncasecmp(str, "to:", 3))
    parse_recips(str+3);
  else if(!strncasecmp(str, "cc:", 3))
    parse_recips(str+3);
  else if(!strncasecmp(str, "bcc:", 4))
    parse_recips(str+4);
}

bool read_header()
{
  mystring line;
  mystring whole;
  bool stripping = false;
  while(fin.getline(line)) {
    if(!line)
      return true;
    if(isspace(line[0]))
      whole += line;
    else {
      parse_line(whole);
      whole = line;
      stripping = false;
      if(!strncasecmp(line.c_str(), "bcc:", 4))
	stripping = true;
    }
    if(!stripping)
      headers.append(line);
  }
  return true;
}

mystring make_messageid()
{
  time_t t = time(0);
  char buf[100];
  strftime(buf, 100, "<%Y.%m.%d.%H.%M.%S.", gmtime(&t));
  return mystring(buf, itoa(getpid()), ".nullmailer@") +
    host_name + ">";
}

mystring make_date()
{
  char buf[256];
  time_t t = time(0);
  strftime(buf, 256, "%a, %d %b %Y %H:%M:%S %Z", localtime(&t));
  return buf;
}

bool fix_header()
{
  if(!has_date)
    headers.append("Date: " + make_date());
  if(!has_from)
    headers.append("From: " + from);
  if(!has_mid)
    headers.append("Message-Id: " + make_messageid());
  return true;
}

///////////////////////////////////////////////////////////////////////////////
// Message sending
///////////////////////////////////////////////////////////////////////////////
static fdobuf* nqpipe = 0;
static pid_t pid = 0;

bool start_queue()
{
  int pipe1[2];
  if(pipe(pipe1) == -1)
    fail_sys("Could not create pipe to nullmailer-queue");
  fout.flush();
  pid = fork();
  if(pid == -1)
    fail_sys("Could not fork");
  if(pid == 0) {
    close(pipe1[1]);
    close(0);
    dup2(pipe1[0], 0);
    execl(QUEUE_BIN, QUEUE_BIN, 0);
    fout.write("Could not exec nullmailer-queue: ");
    fout.writeln(strerror(errno));
    exit(1);
  }
  else {
    close(pipe1[0]);
    nqpipe = new fdobuf(pipe1[1], true);
  }
  return true;
}

bool send_env()
{
  if(!nqpipe->write(sender) || !nqpipe->write("\n"))
    fail("Error sending sender to nullmailer-queue.");
  for(slist::iter iter(recipients); iter; iter++)
    if(!nqpipe->write(*iter) || !nqpipe->write("\n"))
      fail("Error sending recipients to nullmailer-queue.");
  if(!nqpipe->writeln(""))
    fail("Error sending recipients to nullmailer-queue.");
  return true;
}

bool send_header()
{
  for(slist::iter iter(headers); iter; iter++)
    if(!nqpipe->write(*iter) || !nqpipe->write("\n"))
      fail("Error sending header to nullmailer-queue.");
  if(!nqpipe->writeln(""))
    fail("Error sending header to nullmailer-queue.");
  return true;
}

bool send_body()
{
  if(!fdbuf_copy(fin, *nqpipe))
    fail("Error sending message body to nullmailer-queue.");
  return true;
}

bool wait_queue()
{
  if(!nqpipe->close())
    fail("Error closing pipe to nullmailer-queue.");
  int status;
  if(waitpid(pid, &status, 0) == -1)
    fail("Error catching the return value from nullmailer-queue.");
  if(WIFEXITED(status)) {
    status = WEXITSTATUS(status);
    if(status)
      fail("nullmailer-queue failed.");
    else
      return true;
  }
  else
    fail("nullmailer-queue crashed or was killed.");
}

///////////////////////////////////////////////////////////////////////////////
// Main
///////////////////////////////////////////////////////////////////////////////
bool parse_args(int argc, char* argv[])
{
  bool result = true;
  for(int i = 1; i < argc; i++) {
    mystring r = parse_address(argv[i]);
    if(!r) {
      fout.write("Invalid recipient: ");
      fout.writeln(argv[i]);
      result = false;
    }
    else {
      recipients.append(r);
      use_header_recips = false;
    }
  }
  return result;
}

int main(int argc, char* argv[])
{
  set_from();
  if(!parse_args(argc, argv))
    return 1;
  start_queue() && read_header() && fix_header() &&
    send_env() && send_header() && send_body() && wait_queue();
  return 1;
}
